@@ -121,6 +121,17 @@ typedef struct {
 
 const char * kmos_get_license(void) ;
 
+const char * kmos_get_base_name(const char *) ;
+
+cpl_error_code kmos_check_and_set_groups(
+  cpl_frameset *frameset);
+
+int kmos_count_raw_in_frameset(const cpl_frameset *) ;
+
+cpl_frameset * kmos_extract_frameset(
+        const cpl_frameset  *   in,
+        const char          *   tag) ;
+
 cpl_image * kmos_illum_load(
         const char  *   filename,
         cpl_type        im_type,
@@ -133,8 +144,8 @@ char * kmos_get_reflex_suffix(
         const char  *   ifus_txt,
         const char  *   name,
         const char  *   obj_name) ;
-int kmos_collapse_cubes(char *, cpl_frameset *, cpl_parameterlist *, double, 
-        const char *, const char *, double, double, int, int, int) ;
+int kmos_collapse_cubes(const char *, cpl_frameset *, cpl_parameterlist *,
+		double, const char *, const char *, double, double, int, int, int) ;
 cpl_frame * kmos_get_angle_frame(cpl_frameset *, int, const char *) ;
 cpl_frameset * kmos_get_angle_frameset(cpl_frameset *, int, const char *) ;
 cpl_frameset * kmos_purge_wrong_angles_frameset(cpl_frameset *, int, 
@@ -142,6 +153,10 @@ cpl_frameset * kmos_purge_wrong_angles_frameset(cpl_frameset *, int,
 
 int * kmos_get_angles(cpl_frameset *, int *, const char *) ;
 
+int kmos_3dim_clean_plist(cpl_propertylist *) ;
+int kmos_all_clean_plist(cpl_propertylist *) ;
+
+
 cpl_error_code      kmo_cut_endings(cpl_vector** vec,
                                       int *begin,
                                       int *end,
@@ -152,7 +167,7 @@ cpl_error_code      kmo_easy_gaussfit(const cpl_vector *x,
                                       double *sigma,
                                       double *area,
                                       double *offset);
-cpl_vector*         kmo_polyfit_1d(const cpl_vector *x,
+cpl_vector*         kmo_polyfit_1d(   cpl_vector *x,
                                       const cpl_vector *y,
                                       const int degree);
 double              kmo_to_deg(double val);
@@ -180,7 +195,7 @@ cpl_vector*         kmo_idl_values_at_indices(const cpl_vector *data,
                                       const cpl_vector *indices);
 
 /* ---------------------- other --------------------------------------------- */
-cpl_array**         kmo_get_unused_ifus(const cpl_frameset *frameset,
+cpl_array**         kmo_get_unused_ifus(cpl_frameset *frameset,
                                         int omit_telluric,
                                         int omit_illum);
 cpl_error_code      kmo_set_unused_ifus(cpl_array **unused,
