@@ -44,11 +44,6 @@ static char kmo_shift_description[] =
 "left, a positive y-shift shifts upwards, where a shift of one pixel equals\n"
 "0.2arcsec. The output will still have the same dimensions, but the borders \n"
 "will be filled with NaNs accordingly.\n"
-"To adjust only the WCS without moving the data the --wcs-only parameter has to\n"
-"be set to TRUE. The WCS is updated in the same way as if the data would have\n"
-"moved as well. This means that the point at (x,y) has the same coordinates as\n"
-"the point (x+1,y+1) after updating the WCS (the WCS moved in the opposite\n"
-"direction).\n"
 "\n"
 "BASIC PARAMETERS:\n"
 "-----------------\n"
@@ -79,11 +74,6 @@ static char kmo_shift_description[] =
 "turning extrapolation on, the size of the output frame stays the same as for\n"
 "the input frame.\n"
 "\n"
-"--wcs-only\n"
-"By default data and WCS are shifted in sync. If this parameter is set to TRUE\n"
-"only the WCS is updated (i.e. if someone thinks that the IFU isn't pointing\n"
-"exactly to the correct coordinates.)\n"
-"\n"
 "-------------------------------------------------------------------------------\n"
 "  Input files:\n"
 "\n"
@@ -225,17 +215,6 @@ static int kmo_shift_create(cpl_plugin *plugin)
     cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
     cpl_parameterlist_append(recipe->parameters, p);
 
-    /* --wcs-only */
-    p = cpl_parameter_new_value("kmos.kmo_shift.wcs-only",
-                                CPL_TYPE_BOOL,
-                                "FALSE: if data and wcs should be updated."
-                                "TRUE: if only wcs should be corrected",
-                                "kmos.kmo_shift",
-                                0);
-    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "wcs-only");
-    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
-    cpl_parameterlist_append(recipe->parameters, p);
-
     return 0;
 }
 
@@ -307,7 +286,6 @@ static int kmo_shift(cpl_parameterlist *parlist, cpl_frameset *frameset)
                      size                = 0,
                      ifu                 = 0,
                      extrapolate         = 0,
-                     wcs_only            = FALSE,
                      devnr               = 0,
                      index_data          = 0,
                      index_noise         = 0;
@@ -443,14 +421,7 @@ static int kmo_shift(cpl_parameterlist *parlist, cpl_frameset *frameset)
                        CPL_ERROR_ILLEGAL_INPUT,
                        "flux must be TRUE or FALSE!");
 
-        wcs_only = kmo_dfs_get_parameter_bool(parlist, "kmos.kmo_shift.wcs-only");
         KMO_TRY_CHECK_ERROR_STATE();
-        KMO_TRY_EXIT_IF_ERROR(
-            kmo_dfs_print_parameter_help(parlist, "kmos.kmo_shift.wcs-only"));
-
-        KMO_TRY_ASSURE((wcs_only == TRUE) || (wcs_only == FALSE),
-                       CPL_ERROR_NULL_INPUT,
-                       "wcs_only must be TRUE or FALSE!");
 
         cpl_msg_info("", "-------------------------------------------");
 
@@ -539,8 +510,7 @@ static int kmo_shift(cpl_parameterlist *parlist, cpl_frameset *frameset)
                                        &sub_header_data, &sub_header_noise,
                                        pshifts2[2*i-2] / KMOS_PIX_RESOLUTION,
                                        pshifts2[2*i-1] / KMOS_PIX_RESOLUTION,
-                                       flux, devnr, method, extrapol_enum,
-                                       wcs_only));
+                                       flux, method, extrapol_enum));
                 } else {
                     // leave data and noise as they are and
                     // save them again unshifted
