@@ -1,5 +1,4 @@
-/* 
- * This file is part of the KMOS Pipeline
+/* * This file is part of the KMOS Pipeline
  * Copyright (C) 2002,2003 European Southern Observatory
  *
  * This program is free software; you can redistribute it and/or modify
@@ -18,13 +17,18 @@
  */
 
 #ifdef HAVE_CONFIG_H
-#include <config.h>
+  #include <config.h>
 #endif
 
+#define _DEFAULT_SOURCE
+#define _BSD_SOURCE
+
 /*-----------------------------------------------------------------------------
  *                              Includes
  *----------------------------------------------------------------------------*/
 
+#include <stdio.h>
+#include <stdlib.h>
 #include <string.h>
 #include <math.h>
 
@@ -34,16 +38,29 @@
 #include "kmo_debug.h"
 #include "kmo_utils.h"
 #include "kmo_dfs.h"
+#include "kmos_pfits.h"
 #include "kmo_error.h"
 #include "kmo_priv_functions.h"
 #include "kmo_cpl_extensions.h"
 #include "kmo_constants.h"
 #include "kmo_priv_combine.h"
 
+/*-----------------------------------------------------------------------------
+ *                              Define
+ *----------------------------------------------------------------------------*/
+
+#define CPL_DFS_PRO_DID "PRO-1.16"
+
 /*-----------------------------------------------------------------------------
  *                          Functions prototypes
  *----------------------------------------------------------------------------*/
 
+static int  kmos_combine_collect_data(
+        const cpl_propertylist  *   recons_prim_header,
+        const cpl_propertylist  *   recons_ext_header,
+        double                  *   lambda,
+        double                  *   seeing,
+        double                  *   airmass) ;
 static char * kmos_combine_create_ifus_string(const int *,const int *,int);
 static cpl_bivector * kmos_combine_parse_skipped(const char *) ;
 static int kmos_combine_is_skipped(const cpl_bivector *, int, int) ;
@@ -326,44 +343,78 @@ static int kmos_combine(cpl_parameterlist *parlist, cpl_frameset *frameset)
 {
     const cpl_parameter *   par ;
     cpl_frameset        *   rawframes ;
-    const char      * filename,* fmethod,* method,* ifus_txt,* cmethod,
-                    * skipped_frames;
-    double          cpos_rej, cneg_rej ;
-    cpl_vector      *   ifus ;
-    int                 citer, cmin, cmax, nr_frames, index,
-                        data_cube_counter, skip_cube_counter, 
-                        noise_cube_counter, flux, edge_nan, name_vec_size, 
-                        found, suppress_extension, suppress_index, ifu_nr, nv, 
-                        collapse_combined, mapping_id ;
-    char            * tmp_str, * fn_combine, * fn_mask, * mapping_mode,
-                    * name, ** name_vec, * name_loc ;
-    const char      * frame_filename, * tmp_strc ;
-    cpl_image       * exp_mask ;
-    cpl_imagelist   ** data_cube_list, ** noise_cube_list, * cube_combined_data,
-                    * cube_combined_noise ;
-    cpl_propertylist    * main_header, **data_header_list, **noise_header_list,
-                        * tmp_header, * pro_plist ;
+    const char          *   raw_tag,
+                        *   filename,
+                        *   fmethod,
+                        *   method,
+                        *   ifus_txt,
+                        *   cmethod,
+                        *   skipped_frames;
+    double                  cpos_rej, cneg_rej, fwhm_ar_med, fwhm_ne_med,
+                            spec_res, fwhm_ar_val, fwhm_ne_val, pos_ar_val, 
+                            pos_ne_val, airmass, lambda, seeing, sky_res, 
+                            sky_rerr, abmaglim, outer_scale,
+                            crval1, crval2, pos_ar_med, pos_ne_med,
+                            crder3 ;
+    cpl_vector          *   ifus ;
+    int                     citer, cmin, cmax, nr_frames, index, 
+                            data_cube_counter, skip_cube_counter, 
+                            noise_cube_counter, flux, edge_nan, name_vec_size, 
+                            found, suppress_extension, ifu_nr,
+                            nv, collapse_combined, mapping_id, fwhm_count, 
+                            data_count;
+    int                     suppress_index = 0;
+    char                *   tmp_str, 
+                        *   fn_combine, 
+                        *   fn_mask, 
+                        *   mapping_mode,
+                        **  name_vec, 
+                        *   name_loc ;
+    const char          *   name;
+    const char          *   frame_filename, 
+                        *   tmp_strc ;
+    cpl_image           *   exp_mask ;
+    cpl_imagelist       **  data_cube_list, 
+                        **  noise_cube_list, 
+                        *   cube_combined_data,
+                        *   cube_combined_noise ;
+    cpl_propertylist    *   main_header, 
+                        **  data_header_list, 
+                        **  noise_header_list,
+                        *   tmp_header, 
+                        *   plist ;
+    cpl_vector          *   pos_ar ;
+    cpl_vector          *   pos_ne ;
+    cpl_vector          *   fwhm_ar ;
+    cpl_vector          *   fwhm_ne ;
+    cpl_vector          *   sky_res_vec ;
+    cpl_vector          *   airmass_vec ;
+    cpl_vector          *   seeing_vec ;
+    cpl_vector          *   lambda_vec ;
+    cpl_vector          *   tmp_vec ;
     char                *   reflex_suffix ;
-    cpl_bivector    *   skipped_bivector ;
-    cpl_frame       *   frame ;
-    cpl_size            ci ;
-    main_fits_desc      desc;
-    int             *   used_frame_idx ;
-    int             *   used_ifus ;
-    char            *   used_ifus_str ;
-    int             *   skip_frame_idx ;
-    int             *   skip_ifus ;
-    char            *   skip_ifus_str ;
+    cpl_bivector        *   skipped_bivector ;
+    cpl_frame           *   frame ;
+    cpl_size                ci ;
+    main_fits_desc          desc;
+    int                 *   used_frame_idx ;
+    int                 *   used_ifus ;
+    char                *   used_ifus_str ;
+    int                 *   skip_frame_idx ;
+    int                 *   skip_ifus ;
+    char                *   skip_ifus_str ;
     enum extrapolationType  extrapol_enum = NONE_CLIPPING;
-    int             i, j ;
+    double                  light_speed ;
+    int                     i, j ;
 
-    /* Check entries */
-    if (parlist == NULL || frameset == NULL) {
-        cpl_msg_error(__func__, "Null Inputs") ;
-        cpl_error_set(__func__, CPL_ERROR_NULL_INPUT) ;
-        return -1 ;
+    /* Check initial Entries */
+    if (kmos_check_and_set_groups(frameset) != CPL_ERROR_NONE) {
+        return cpl_error_get_code();
     }
 
+    /* Initialise */
+    light_speed = 299792.48 ;
+
     /* Get parameters */
     par = cpl_parameterlist_find_const(parlist, "kmos.kmos_combine.method");
     method = cpl_parameter_get_string(par) ;
@@ -374,7 +425,7 @@ static int kmos_combine(cpl_parameterlist *parlist, cpl_frameset *frameset)
     par = cpl_parameterlist_find_const(parlist, "kmos.kmos_combine.ifus");
     ifus_txt = cpl_parameter_get_string(par) ;
     par = cpl_parameterlist_find_const(parlist, "kmos.kmos_combine.name");
-    name = (char*)cpl_parameter_get_string(par) ;
+    name = cpl_parameter_get_string(par) ;
     par = cpl_parameterlist_find_const(parlist, "kmos.kmos_combine.flux");
     flux = cpl_parameter_get_bool(par) ;
     par = cpl_parameterlist_find_const(parlist, 
@@ -412,13 +463,6 @@ static int kmos_combine(cpl_parameterlist *parlist, cpl_frameset *frameset)
         return -1 ;
     }
 
-    /* Identify the RAW and CALIB frames in the input frameset */
-    if (kmo_dfs_set_groups(frameset) != 1) {
-        cpl_msg_error(__func__, "Cannot identify RAW and CALIB frames") ;
-        cpl_error_set(__func__, CPL_ERROR_ILLEGAL_INPUT) ;
-        return -1 ;
-    }
-
     /* Parse the skipped frames option */
     skipped_bivector = NULL ;
     if (strcmp(skipped_frames, "")) {
@@ -442,6 +486,10 @@ static int kmos_combine(cpl_parameterlist *parlist, cpl_frameset *frameset)
         return -1 ;
     }
 
+    /* Get the rawframes tag used */
+    frame = cpl_frameset_get_position(rawframes, 0);
+    raw_tag = cpl_frame_get_tag(frame) ;
+
     /* Load IFUS if specified */
     if (strcmp(ifus_txt, "")) {
         ifus = kmo_identify_values(ifus_txt);
@@ -570,7 +618,6 @@ static int kmos_combine(cpl_parameterlist *parlist, cpl_frameset *frameset)
             sizeof(cpl_imagelist*));
     noise_header_list = cpl_calloc(nr_frames*KMOS_NR_IFUS,
             sizeof(cpl_propertylist*));
- 
     /*
     if (ifus != NULL) cpl_vector_delete(ifus);
     cpl_frameset_delete(rawframes) ;
@@ -588,6 +635,13 @@ static int kmos_combine(cpl_parameterlist *parlist, cpl_frameset *frameset)
     for (nv = 0; nv < name_vec_size; nv++) {
         name_loc = name_vec[nv];
 
+        fwhm_ar = cpl_vector_new(nr_frames*KMOS_NR_IFUS) ;
+        fwhm_ne = cpl_vector_new(nr_frames*KMOS_NR_IFUS) ;
+        pos_ar = cpl_vector_new(nr_frames*KMOS_NR_IFUS) ;
+        pos_ne = cpl_vector_new(nr_frames*KMOS_NR_IFUS) ;
+        airmass_vec = cpl_vector_new(nr_frames*KMOS_NR_IFUS) ;
+        seeing_vec = cpl_vector_new(nr_frames*KMOS_NR_IFUS) ;
+        lambda_vec = cpl_vector_new(nr_frames*KMOS_NR_IFUS) ;
         used_frame_idx = cpl_calloc(nr_frames*KMOS_NR_IFUS, sizeof(int)) ;
         used_ifus = cpl_calloc(nr_frames*KMOS_NR_IFUS, sizeof(int)) ;
         skip_frame_idx = cpl_calloc(nr_frames*KMOS_NR_IFUS, sizeof(int)) ;
@@ -596,12 +650,15 @@ static int kmos_combine(cpl_parameterlist *parlist, cpl_frameset *frameset)
         data_cube_counter = 0;
         skip_cube_counter = 0;
         noise_cube_counter = 0;
+        fwhm_count = 0 ;
+        data_count = 0 ;
         for (i = 0; i < nr_frames; i++) {
             tmp_str = cpl_sprintf("%d", i);
             frame = kmo_dfs_get_frame(rawframes, tmp_str);
             frame_filename = cpl_frame_get_filename(frame);
             kmo_init_fits_desc(&desc);
             desc = kmo_identify_fits_header(frame_filename);
+            main_header = cpl_propertylist_load(frame_filename, 0) ;
 
             if (mapping_mode != NULL) {
                 /* Mapping mode */
@@ -622,6 +679,44 @@ static int kmos_combine(cpl_parameterlist *parlist, cpl_frameset *frameset)
                             data_header_list[data_cube_counter] =
                                 kmo_dfs_load_sub_header(rawframes, 
                                         tmp_str, j, FALSE);
+
+                            /* Store some other quantities */
+                            if (kmos_combine_collect_data(main_header,
+                                    data_header_list[data_cube_counter],
+                                    &lambda, &seeing, &airmass) == -1) {
+                                cpl_error_reset() ;
+                            } else {
+                                cpl_vector_set(lambda_vec, data_count, lambda) ;
+                                cpl_vector_set(seeing_vec, data_count, seeing) ;
+                                cpl_vector_set(airmass_vec,data_count,airmass) ;
+                                data_count ++ ;
+                            }
+
+                            /* Store the Resolution */
+                            fwhm_ar_val = kmos_pfits_get_qc_ar_fwhm_mean(
+                                    data_header_list[data_cube_counter]) ;
+                            fwhm_ne_val = kmos_pfits_get_qc_ne_fwhm_mean(
+                                    data_header_list[data_cube_counter]) ;
+                            pos_ar_val = kmos_pfits_get_qc_ar_pos_stdev(
+                                    data_header_list[data_cube_counter]) ;
+                            pos_ne_val = kmos_pfits_get_qc_ne_pos_stdev(
+                                    data_header_list[data_cube_counter]) ;
+                            if (cpl_error_get_code() != CPL_ERROR_NONE) {
+                                cpl_msg_warning(__func__, 
+                                        "Cannot retrieve QC ARC xx FWHM MEAN") ;
+                                cpl_error_reset() ;
+                            } else {
+                                cpl_vector_set(fwhm_ar,fwhm_count,
+                                        light_speed/fwhm_ar_val) ;
+                                cpl_vector_set(fwhm_ne,fwhm_count,
+                                        light_speed/fwhm_ne_val) ;
+                                cpl_vector_set(pos_ar,fwhm_count,
+                                        pos_ar_val*lambda*1000/light_speed) ;
+                                cpl_vector_set(pos_ne,fwhm_count,
+                                        pos_ne_val*lambda*1000/light_speed) ;
+                                fwhm_count++ ;
+                            }
+
                             cpl_propertylist_update_string(
                                     data_header_list[data_cube_counter],
                                     "ESO PRO FRNAME", frame_filename);
@@ -697,6 +792,44 @@ static int kmos_combine(cpl_parameterlist *parlist, cpl_frameset *frameset)
                             data_header_list[data_cube_counter] =
                                 kmo_dfs_load_sub_header(rawframes, tmp_str,
                                         ifu_nr, FALSE);
+
+                            /* Store some other quantities */
+                            if (kmos_combine_collect_data(main_header,
+                                    data_header_list[data_cube_counter],
+                                    &lambda, &seeing, &airmass) == -1) {
+                                cpl_error_reset() ;
+                            } else {
+                                cpl_vector_set(lambda_vec, data_count, lambda) ;
+                                cpl_vector_set(seeing_vec, data_count, seeing) ;
+                                cpl_vector_set(airmass_vec,data_count,airmass) ;
+                                data_count ++ ;
+                            }
+
+                            /* Store the Resolution and crder3 */
+                            fwhm_ar_val = kmos_pfits_get_qc_ar_fwhm_mean(
+                                    data_header_list[data_cube_counter]) ;
+                            fwhm_ne_val = kmos_pfits_get_qc_ne_fwhm_mean(
+                                    data_header_list[data_cube_counter]) ;
+                            pos_ar_val = kmos_pfits_get_qc_ar_pos_stdev(
+                                    data_header_list[data_cube_counter]) ;
+                            pos_ne_val = kmos_pfits_get_qc_ne_pos_stdev(
+                                    data_header_list[data_cube_counter]) ;
+                            if (cpl_error_get_code() != CPL_ERROR_NONE) {
+                                cpl_msg_warning(__func__, 
+                                        "Cannot retrieve QC ARC xx FWHM MEAN") ;
+                                cpl_error_reset() ;
+                            } else {
+                                cpl_vector_set(fwhm_ar,fwhm_count,
+                                        light_speed/fwhm_ar_val) ;
+                                cpl_vector_set(fwhm_ne,fwhm_count,
+                                        light_speed/fwhm_ne_val) ;
+                                cpl_vector_set(pos_ar,fwhm_count,
+                                        pos_ar_val*lambda*1000/light_speed) ;
+                                cpl_vector_set(pos_ne,fwhm_count,
+                                        pos_ne_val*lambda*1000/light_speed) ;
+                                fwhm_count++ ;
+                            }
+
                             cpl_propertylist_update_string(
                                     data_header_list[data_cube_counter],
                                     "ESO PRO FRNAME", frame_filename);
@@ -743,10 +876,76 @@ static int kmos_combine(cpl_parameterlist *parlist, cpl_frameset *frameset)
                     }
                 } 
             }
+            cpl_propertylist_delete(main_header) ;
             kmo_free_fits_desc(&desc);
             cpl_free(tmp_str);
         } 
- 
+
+        /* Compute the Resolution */
+        if (fwhm_count > 2) {
+            tmp_vec = cpl_vector_extract(fwhm_ar, 0, fwhm_count-1, 1) ;
+            fwhm_ar_med = cpl_vector_get_median(tmp_vec) ;
+            cpl_vector_delete(tmp_vec) ;
+            tmp_vec = cpl_vector_extract(fwhm_ne, 0, fwhm_count-1, 1) ;
+            fwhm_ne_med = cpl_vector_get_median(tmp_vec) ;
+            cpl_vector_delete(tmp_vec) ;
+            tmp_vec = cpl_vector_extract(pos_ar, 0, fwhm_count-1, 1) ;
+            pos_ar_med = cpl_vector_get_median(tmp_vec) ;
+            cpl_vector_delete(tmp_vec) ;
+            tmp_vec = cpl_vector_extract(pos_ne, 0, fwhm_count-1, 1) ;
+            pos_ne_med = cpl_vector_get_median(tmp_vec) ;
+            cpl_vector_delete(tmp_vec) ;
+            spec_res = (fwhm_ar_med + fwhm_ne_med) / 2.0 ;
+            crder3 = (pos_ar_med + pos_ne_med) / 2.0 ;
+        } else if (fwhm_count > 0) {
+            tmp_vec = cpl_vector_extract(fwhm_ar, 0, fwhm_count-1, 1) ;
+            fwhm_ar_med = cpl_vector_get_mean(tmp_vec) ;
+            cpl_vector_delete(tmp_vec) ;
+            tmp_vec = cpl_vector_extract(fwhm_ne, 0, fwhm_count-1, 1) ;
+            fwhm_ne_med = cpl_vector_get_mean(tmp_vec) ;
+            cpl_vector_delete(tmp_vec) ;
+            tmp_vec = cpl_vector_extract(pos_ar, 0, fwhm_count-1, 1) ;
+            pos_ar_med = cpl_vector_get_mean(tmp_vec) ;
+            cpl_vector_delete(tmp_vec) ;
+            tmp_vec = cpl_vector_extract(pos_ne, 0, fwhm_count-1, 1) ;
+            pos_ne_med = cpl_vector_get_mean(tmp_vec) ;
+            cpl_vector_delete(tmp_vec) ;
+            spec_res = (fwhm_ar_med + fwhm_ne_med) / 2.0 ;
+            crder3 = (pos_ar_med + pos_ne_med) / 2.0 ;
+        } else {
+            spec_res = -1.0 ;
+            crder3 = -1.0 ;
+        }
+        cpl_vector_delete(fwhm_ar) ;
+        cpl_vector_delete(fwhm_ne) ;
+        cpl_vector_delete(pos_ar) ;
+        cpl_vector_delete(pos_ne) ;
+
+        /* Compute SKY_RES / SKY_RERR  */
+        sky_res_vec = cpl_vector_new(data_count) ;
+        for (i=0 ; i<data_count ; i++) {
+            lambda = cpl_vector_get(lambda_vec, i) ;
+            seeing = cpl_vector_get(seeing_vec, i) ;
+            airmass = cpl_vector_get(airmass_vec, i) ;
+            outer_scale=sqrt(1-78.08*((pow(lambda*1e-6,0.4)*pow(airmass,-0.2))/
+                        (pow(seeing,1./3.)))) ;
+            sky_res = seeing * pow(0.5/lambda, 0.2) * pow(airmass, 3./5.) * 
+                outer_scale ;
+            cpl_vector_set(sky_res_vec, i, sky_res) ;
+        }
+        cpl_vector_delete(airmass_vec) ;
+        cpl_vector_delete(seeing_vec) ;
+        cpl_vector_delete(lambda_vec) ;
+
+        if (data_count > 2) {
+            sky_res = cpl_vector_get_median_const(sky_res_vec) ;
+            sky_rerr = sqrt((0.2*0.2)+pow(cpl_vector_get_stdev(sky_res_vec),2));
+        } else {
+            sky_res = cpl_vector_get_mean(sky_res_vec) ;
+            sky_rerr = 0.2;
+        }
+        cpl_vector_delete(sky_res_vec) ;
+
         /* Create String for the output header - IFUs usage */
         used_ifus_str = kmos_combine_create_ifus_string(used_frame_idx, 
                 used_ifus, data_cube_counter);
@@ -815,6 +1014,11 @@ static int kmos_combine(cpl_parameterlist *parlist, cpl_frameset *frameset)
             cpl_propertylist_delete(noise_header_list[0]) ;
         cpl_frameset_delete(rawframes) ;
         */
+
+        /* Compute ABMAGLIM */
+        abmaglim = kmos_idp_compute_abmaglim(cube_combined_data,
+                sky_res, data_header_list[0]) ;
+
         /* Save data */
         if (!suppress_extension) {
             /* Setup output category COMBINE + ESO PRO CATG */
@@ -832,28 +1036,70 @@ static int kmos_combine(cpl_parameterlist *parlist, cpl_frameset *frameset)
         }
 
         /* Create PRO keys plist */
-        pro_plist = cpl_propertylist_new() ;
+        plist = cpl_propertylist_new() ;
         if (used_ifus_str != NULL) 
-            cpl_propertylist_update_string(pro_plist, "ESO PRO USEDIFUS", 
+            cpl_propertylist_update_string(plist, "ESO PRO USEDIFUS", 
                     used_ifus_str) ;
         if (skip_ifus_str != NULL) 
-            cpl_propertylist_update_string(pro_plist, "ESO PRO SKIPPEDIFUS", 
+            cpl_propertylist_update_string(plist, "ESO PRO SKIPPEDIFUS", 
                     skip_ifus_str) ;
 
+        /* QC EXPMASK AVG */
+        cpl_propertylist_update_double(plist, "ESO QC EXPMASK AVG",
+                cpl_image_get_mean(exp_mask)) ;
+
+		/* PIXNOISE */
+		cpl_propertylist_update_double(plist, KEY_PIXNOISE, 
+                cpl_image_get_median(
+                    cpl_imagelist_get_const(cube_combined_data, 1000))) ;
+        cpl_propertylist_set_comment(plist, KEY_PIXNOISE, KEY_PIXNOISE_COMMENT);
+
+        /* Store SPEC_REC in the header */
+        cpl_propertylist_update_double(plist, KEY_SPEC_RES, spec_res) ;
+        cpl_propertylist_set_comment(plist, KEY_SPEC_RES, KEY_SPEC_RES_COMMENT);
+
+        /* Store data_cube_counter */
+        cpl_propertylist_update_int(plist, "ESO QC COMBINED_CUBES NB",
+                data_cube_counter) ;
+
+        /* EXPMASK NAME */
+        tmp_str = cpl_sprintf("%s.fits", fn_mask) ;
+		cpl_propertylist_update_string(plist, "ESO QC EXPMASK NAME", tmp_str) ;
+        cpl_free(tmp_str) ;
+
+        /* Add the Collapsed File Name */
+        if (collapse_combined) {
+            tmp_str = cpl_sprintf("make_image_%s", fn_combine) ;
+            cpl_propertylist_update_string(plist, "ESO QC COLLAPSE NAME", 
+                    tmp_str) ;
+            cpl_free(tmp_str) ;
+        } 
+
+		/* ABMAGLIM */
+		cpl_propertylist_update_double(plist, KEY_ABMAGLIM, abmaglim) ;
+        cpl_propertylist_set_comment(plist, KEY_ABMAGLIM, KEY_ABMAGLIM_COMMENT);
+
+		/* SKY_RES / SKY_RERR */
+		cpl_propertylist_update_double(plist, KEY_SKY_RES, sky_res) ;
+        cpl_propertylist_set_comment(plist, KEY_SKY_RES, KEY_SKY_RES_COMMENT) ;
+		cpl_propertylist_update_double(plist, KEY_SKY_RERR, sky_rerr) ;
+        cpl_propertylist_set_comment(plist, KEY_SKY_RERR,KEY_SKY_RERR_COMMENT) ;
+
         /* Add REFLEX SUFFIX keyword */
         reflex_suffix = kmos_get_reflex_suffix(mapping_id,
                 ifus_txt, name, name_loc) ;
-        cpl_propertylist_update_string(pro_plist, "ESO PRO REFLEX SUFFIX", 
+        cpl_propertylist_update_string(plist, "ESO PRO REFLEX SUFFIX", 
                 reflex_suffix) ;
         cpl_free(reflex_suffix) ;
 
         /* Save Headers first */
         frame = cpl_frameset_find(rawframes, NULL);
-        kmo_dfs_save_main_header(frameset, fn_combine, "", frame, pro_plist, 
+        kmo_dfs_save_main_header(frameset, fn_combine, "", frame, plist, 
                 parlist, cpl_func);
-        kmo_dfs_save_main_header(frameset, fn_mask, "", frame, pro_plist, 
+
+        kmo_dfs_save_main_header(frameset, fn_mask, "", frame, plist, 
                 parlist, cpl_func);
-        cpl_propertylist_delete(pro_plist) ;
+        cpl_propertylist_delete(plist) ;
         cpl_free(used_ifus_str) ;
         cpl_free(skip_ifus_str) ;
 
@@ -874,24 +1120,27 @@ static int kmos_combine(cpl_parameterlist *parlist, cpl_frameset *frameset)
                 cpl_propertylist_erase(noise_header_list[0], "ESO PRO IFUNR");
             }
         }
-        kmo_dfs_save_cube(cube_combined_data, fn_combine, "", 
-                data_header_list[0], 0./0.);
+        /* Save DATA extension */
+        plist = cpl_propertylist_duplicate(data_header_list[0]) ;
+
+        /* Store CRDER3 in the header */
+        cpl_propertylist_update_double(plist, KEY_CRDER3, crder3) ;
+        cpl_propertylist_set_comment(plist, KEY_CRDER3, KEY_CRDER3_COMMENT);
+
+        kmo_dfs_save_cube(cube_combined_data, fn_combine, "", plist, 0./0.);
+        cpl_propertylist_delete(plist) ;
         cpl_imagelist_delete(cube_combined_data);
-        kmo_dfs_save_cube(cube_combined_noise, fn_combine, "", 
-                noise_header_list[0], 0./0.);
+
+        /* Save NOISE extension */
+        plist = cpl_propertylist_duplicate(noise_header_list[0]) ;
+        kmo_dfs_save_cube(cube_combined_noise, fn_combine, "", plist, 0./0.);
         cpl_imagelist_delete(cube_combined_noise);
+        cpl_propertylist_delete(plist) ;
         cpl_free(fn_combine);
         
-        cpl_propertylist_erase(data_header_list[0], CRPIX3);
-        cpl_propertylist_erase(data_header_list[0], CRPIX3);
-        cpl_propertylist_erase(data_header_list[0], CRVAL3);
-        cpl_propertylist_erase(data_header_list[0], CDELT3);
-        cpl_propertylist_erase(data_header_list[0], CD1_3);
-        cpl_propertylist_erase(data_header_list[0], CD2_3);
-        cpl_propertylist_erase(data_header_list[0], CD3_1);
-        cpl_propertylist_erase(data_header_list[0], CD3_2);
-        cpl_propertylist_erase(data_header_list[0], CD3_3);
-        cpl_propertylist_erase(data_header_list[0], CTYPE3);
+        /* Clean 3rd dimension keys */
+        kmos_3dim_clean_plist(data_header_list[0]) ;
+
         kmo_dfs_save_image(exp_mask, fn_mask, "", data_header_list[0], 0./0.);
         cpl_free(fn_mask);
         cpl_image_delete(exp_mask);
@@ -903,7 +1152,6 @@ static int kmos_combine(cpl_parameterlist *parlist, cpl_frameset *frameset)
         if (noise_header_list!=NULL && noise_cube_counter==0) 
             cpl_propertylist_delete(noise_header_list[0]) ;
     } 
-    cpl_frameset_delete(rawframes) ;
     if (skipped_bivector!=NULL) cpl_bivector_delete(skipped_bivector) ;
     if (ifus != NULL) cpl_vector_delete(ifus);
     cpl_free(data_cube_list);
@@ -916,10 +1164,138 @@ static int kmos_combine(cpl_parameterlist *parlist, cpl_frameset *frameset)
 
     /* Collapse the combined cubes if requested */
     if (collapse_combined) {
-        kmos_collapse_cubes(COMBINED_RECONS, frameset, parlist, 0.1, "",
-                DEF_REJ_METHOD, DEF_POS_REJ_THRES, DEF_NEG_REJ_THRES,
-                DEF_ITERATIONS, DEF_NR_MIN_REJ, DEF_NR_MAX_REJ) ;
+    	kmos_collapse_cubes(COMBINED_CUBE, frameset, parlist, 0.1, "",
+    	                DEF_REJ_METHOD, DEF_POS_REJ_THRES, DEF_NEG_REJ_THRES,
+    	                DEF_ITERATIONS, DEF_NR_MIN_REJ, DEF_NR_MAX_REJ);
+
     }
+
+    /* Create the IDP Products */
+    cpl_frame * combined_frame ;
+    const char * combined_fname ;
+    char * idp_combined_fname ;
+    cpl_imagelist * cube_combined_error ;
+    const char * const_data_extname ;
+    char * data_extname ;
+    char * error_extname ;
+    char * object ;
+    cpl_frameset * combined_frames ;
+
+    /* Loop on the Combined Frames */
+    combined_frames = kmos_extract_frameset(frameset,       COMBINED_CUBE) ;
+    combined_frame  = kmo_dfs_get_frame(combined_frames,    COMBINED_CUBE);
+    while (combined_frame != NULL ) {
+        combined_fname = cpl_frame_get_filename(combined_frame);
+
+        /* Load combined data cube */
+        cube_combined_data=cpl_imagelist_load(combined_fname, CPL_TYPE_FLOAT,1);
+
+        /* Compute the IDP error */
+        cube_combined_error = kmos_idp_compute_error(cube_combined_data);
+
+        /************* IDP  PRIMARY  **********/
+        main_header =   cpl_propertylist_load(combined_fname, 0) ;
+        plist =         cpl_propertylist_load(combined_fname, 1) ;
+        kmos_idp_prepare_main_keys(main_header, frameset, plist,
+                raw_tag, cube_combined_error) ;
+        cpl_propertylist_delete(plist) ;
+        frame = cpl_frameset_find(rawframes, NULL);
+
+        /* Save Primary */
+        idp_combined_fname = cpl_sprintf("IDP_%s", combined_fname) ;
+
+        /***************************************************/
+        /* Replace the call of cpl_dfs_save_propertylist() */
+        /* because RA/DEC need to be changed               */
+        /***************************************************/
+        const char * procat = cpl_propertylist_get_string(main_header, 
+                CPL_DFS_PRO_CATG);
+        cpl_msg_info(cpl_func, "Writing FITS propertylist product(%s): %s", 
+                procat, idp_combined_fname);
+
+        cpl_frame * product_frame = cpl_frame_new();
+        cpl_frame_set_filename(product_frame, idp_combined_fname);
+        cpl_frame_set_tag(product_frame, procat);
+        cpl_frame_set_type(product_frame, CPL_FRAME_TYPE_ANY);
+        cpl_frame_set_group(product_frame, CPL_FRAME_GROUP_PRODUCT);
+        cpl_frame_set_level(product_frame, CPL_FRAME_LEVEL_FINAL);
+
+        plist = cpl_propertylist_duplicate(main_header);
+
+        /* Add DataFlow keywords */
+        cpl_dfs_setup_product_header(plist, product_frame, frameset, parlist, 
+                cpl_func, VERSION, CPL_DFS_PRO_DID, frame);
+
+        /* Fix RA / DEC <- Set to CRVAL1 / CRVAL2 */
+        /* plist(RA / DEC) from plist2(CRVAL1 / CRVAL2) */
+        cpl_propertylist * plist2 = cpl_propertylist_load(combined_fname, 1) ;
+        crval1 = cpl_propertylist_get_double(plist2, CRVAL1);
+        crval2 = cpl_propertylist_get_double(plist2, CRVAL2);
+        cpl_propertylist_delete(plist2) ;
+        kmclipm_update_property_double(plist, "RA", crval1, "") ;
+        kmclipm_update_property_double(plist, "DEC", crval2, "") ;
+
+        /* Save and register */
+        cpl_propertylist_save(plist, idp_combined_fname, CPL_IO_CREATE);
+        cpl_propertylist_delete(plist) ;
+        cpl_frameset_insert(frameset, product_frame);
+        /* cpl_frame_delete(product_frame); */
+
+        /***************************************************/
+        /* Instead of :                                    */
+        /***************************************************/
+        /*
+        cpl_dfs_save_propertylist(frameset, NULL, parlist, frameset, frame,
+                cpl_func, main_header, NULL, VERSION, idp_combined_fname);
+        */
+        /***************************************************/
+
+        /* Remember OBJECT for the extensions */
+        object = cpl_strdup(kmos_pfits_get_object(main_header)) ;
+
+        cpl_propertylist_delete(main_header) ;
+
+        /************* IDP DATA extension **********/
+        /* Handle extname values */
+        plist =         cpl_propertylist_load(combined_fname, 1) ;
+        const_data_extname = kmos_pfits_get_extname(plist) ;
+        error_extname = kmos_idp_compute_error_extname(const_data_extname) ;
+
+        /* IDP Add DATA Keywords */
+        kmos_idp_prepare_data_keys(plist, error_extname, object) ;
+        cpl_free(error_extname) ;
+
+        /* Ѕave Data */
+        cpl_imagelist_save(cube_combined_data, idp_combined_fname, 
+                CPL_BPP_IEEE_FLOAT,
+                plist, CPL_IO_EXTEND) ;
+        cpl_propertylist_delete(plist) ;
+        cpl_imagelist_delete(cube_combined_data);
+
+        /************* IDP ERROR extension **********/
+        /* IDP Add ERROR Keywords */
+        plist = cpl_propertylist_load(combined_fname, 1) ;
+        kmclipm_update_property_string(plist, EXTNAME, "DATA", "");
+        data_extname = cpl_strdup(kmos_pfits_get_extname(plist)) ;
+        error_extname = kmos_idp_compute_error_extname(data_extname) ;
+        kmos_idp_prepare_error_keys(plist, error_extname, data_extname, object);
+        cpl_free(object) ;
+        cpl_free(error_extname) ;
+        cpl_free(data_extname) ;
+
+        /* Save ERROR extension */
+        cpl_imagelist_save(cube_combined_error, idp_combined_fname, 
+                CPL_BPP_IEEE_FLOAT,
+                plist, CPL_IO_EXTEND) ;
+        cpl_imagelist_delete(cube_combined_error);
+        cpl_propertylist_delete(plist) ;
+        cpl_free(idp_combined_fname) ;
+
+        /* Next candidate */
+        combined_frame = kmo_dfs_get_frame(combined_frames, NULL);
+    }
+    cpl_frameset_delete(rawframes) ;
+    cpl_frameset_delete(combined_frames) ;
     return 0;
 }
 
@@ -969,6 +1345,54 @@ static char * kmos_combine_create_ifus_string(
     return cpl_strdup(out) ;
 }
 
+/*----------------------------------------------------------------------------*/
+/**
+  @brief 
+  @param  
+  @return 
+ */
+/*----------------------------------------------------------------------------*/
+static int  kmos_combine_collect_data(
+        const cpl_propertylist  *   recons_prim_header,
+        const cpl_propertylist  *   recons_ext_header,
+        double                  *   lambda,
+        double                  *   seeing,
+        double                  *   airmass)
+{
+
+    double  crval3, cd3_3, crpix3, airm_start, airm_end, fwhmlin ;
+    int     naxis3 ;
+    
+    /* Check Entries */
+    if (recons_prim_header == NULL || recons_ext_header == NULL) return -1;
+    if (lambda == NULL || seeing == NULL || airmass == NULL) return -1;
+
+    /* Get Values from header */
+	crval3 =        kmos_pfits_get_crval3(recons_ext_header) ;
+    cd3_3 =         kmos_pfits_get_cd3_3(recons_ext_header) ;
+    crpix3 =        kmos_pfits_get_crpix3(recons_ext_header) ;
+    naxis3 =        kmos_pfits_get_naxis3(recons_ext_header) ;
+    airm_start =    kmos_pfits_get_airmass_start(recons_prim_header) ;
+    airm_end =      kmos_pfits_get_airmass_end(recons_prim_header) ;
+    fwhmlin =       kmos_pfits_get_ia_fwhmlin(recons_prim_header) ;
+    if (cpl_error_get_code() != CPL_ERROR_NONE) {
+        cpl_error_reset() ;
+        cpl_msg_error(__func__, "Cannot collect data from header") ;
+        return -1 ;
+    }
+
+    /* Lambda */
+    *lambda = (crval3 + (crval3 + cd3_3 * naxis3 - (crpix3-1) * cd3_3)) / 2.0;
+
+    /* Airmass */
+    *airmass = (airm_start + airm_end) / 2.0 ;
+
+    /* Seeing */
+    *seeing = fwhmlin ;
+
+    return 0 ;
+}
+
 /*----------------------------------------------------------------------------*/
 /**
   @brief 
@@ -985,8 +1409,8 @@ static cpl_bivector * kmos_combine_parse_skipped(const char * str)
     int                 nb_values ;
     char            *   s1 ;
     char            *   s2 ;
-    int                 val1, val2, ret ;
-    
+    int                 val1, val2;
+
     /* Check Entries */
     if (str == NULL) return NULL ;
 
